@@ -1,11 +1,14 @@
 import builtins
+import copy
 import functools
 import itertools
+import math
 import operator
 import re
 import sys
 import typing
-from collections import Counter, deque
+from collections import Counter, UserList, deque
+from heapq import heapify, heappop, heappush
 
 from typing_extensions import ParamSpec
 
@@ -46,24 +49,11 @@ def chunk_default(
     )
 
 
-class list(builtins.list, typing.Generic[T]):
+class list(UserList, typing.Generic[T]):
     """Smart/fluent list class"""
 
     _SENTINEL = object()
 
-    @typing.overload
-    def __getitem__(self, index: int) -> T:
-        ...
-
-    @typing.overload
-    def __getitem__(self, index: slice) -> "list[T]":
-        ...
-
-    def __getitem__(self, index):
-        if isinstance(index, int):
-            return super().__getitem__(index)
-        return list(super().__getitem__(index))
-
     def iter(self) -> "iter[T]":
         """Return an iterator over the list."""
         return iter(self)
@@ -75,39 +65,47 @@ class list(builtins.list, typing.Generic[T]):
         return list(map(func, self))
 
     def filtered(
-        self, pred: typing.Union[typing.Callable[[T], bool], None]
+        self, pred: typing.Union[typing.Callable[[T], bool], T, None] = None
     ) -> "list[T]":
         """Return a list containing only the elements for which pred
         returns True.
 
         If pred is None, return a list containing only elements that are
         truthy.
+
+        If pred is a T (and T is not a callable or None), return a list
+        containing only elements that compare equal to pred.
         """
+        if not callable(pred) and pred is not None:
+            pred = (lambda j: lambda i: i == j)(pred)
         return list(filter(pred, self))
 
-    def _window(
-        self, window_size: int
-    ) -> typing.Generator[typing.Tuple[T, ...], None, None]:
-        elements = deque()
-        for _ in range(window_size):
-            try:
-                elements.append(self.next())
-            except StopIteration:
-                return
+    def find(
+        self, pred: typing.Union[typing.Callable[[T], bool], T, None] = None
+    ) -> typing.Optional[T]:
+        """Return the first element of self for which pred returns True.
 
-        yield tuple(elements)
+        If pred is None, return the first element which is truthy.
 
-        for el in self:
-            elements.popleft()
-            elements.append(el)
-            yield tuple(elements)
+        If pred is a T (and T is not a callable or None), return the first element
+        that compares equal to pred.
+
+        If no such element exists, return None.
+        """
+        if pred is None:
+            pred = lambda i: bool(i)
+        elif not callable(pred):
+            pred = (lambda j: lambda i: i == j)(pred)
+        for i in self:
+            if pred(i):
+                return i
 
     def windowed(self, window_size: int) -> "list[typing.Tuple[T, ...]]":
         """Return an list containing the elements of this list in
         a sliding window of size window_size. If there are not enough elements
         to create a full window, the list will be empty.
         """
-        return list(self._window(window_size))
+        return list(self.iter().window(window_size))
 
     def shifted_zip(self, shift: int = 1) -> "iter[typing.Tuple[T, ...]]":
         """Return an iterator containing pairs of elements separated by shift.
@@ -125,7 +123,7 @@ class list(builtins.list, typing.Generic[T]):
         ...
 
     def reduce(self, func, initial=_SENTINEL):
-        """Reduce the iterator to a single value, using the reduction
+        """Reduce the list to a single value, using the reduction
         function provided.
         """
         if initial is self._SENTINEL:
@@ -150,13 +148,13 @@ class list(builtins.list, typing.Generic[T]):
 
     def accumulated(self, func=operator.add, initial=_SENTINEL):
         """Return the accumulated results of calling func on the elements in
-        this iterator.
+        this list.
 
         initial is only usable on versions of Python equal to or greater than 3.8.
         """
         if initial is self._SENTINEL:
-            return list(itertools.accumulate(func, self))
-        return list(itertools.accumulate(func, self, initial))
+            return list(itertools.accumulate(self, func))
+        return list(itertools.accumulate(self, func, initial))
 
     def chunked(self, n: int) -> "list[typing.Tuple[T, ...]]":
         """Return a list containing the elements of this list in chunks
@@ -189,6 +187,23 @@ class list(builtins.list, typing.Generic[T]):
             return sum(self)
         return sum(self, initial)
 
+    @typing.overload
+    def prod(self) -> T:
+        ...
+
+    @typing.overload
+    def prod(self, initial: T) -> T:
+        ...
+
+    def prod(self, initial=_SENTINEL):
+        """Return the product of all elements in this list.
+
+        If initial is provided, it is used as the initial value.
+        """
+        if initial is self._SENTINEL:
+            return math.prod(self)
+        return math.prod(self, start=initial)
+
     def sorted(
         self,
         key: typing.Union[typing.Callable[[T], U], None] = None,
@@ -280,8 +295,17 @@ class list(builtins.list, typing.Generic[T]):
             )
         )
 
+    def enumerated(self, start: int = 0) -> "list[typing.Tuple[int, T]]":
+        return list(enumerate(self, start))
 
-class iter(typing.Generic[T]):
+    def deepcopy(self) -> "list[T]":
+        return copy.deepcopy(self)
+
+    def __repr__(self) -> str:
+        return f"list({super().__repr__()})"
+
+
+class iter(typing.Generic[T], typing.Iterator[T], typing.Iterable[T]):
     """Smart/fluent iterator class"""
 
     _SENTINEL = object()
@@ -292,21 +316,51 @@ class iter(typing.Generic[T]):
     def __iter__(self) -> typing.Iterable[T]:
         return self.it.__iter__()
 
+    def __next__(self) -> T:
+        return next(self)
+
     def map(self, func: typing.Callable[[T], U]) -> "iter[U]":
         """Return an iterator containing the result of calling func on each
         element in this iterator.
         """
         return iter(map(func, self))
 
-    def filter(self, pred: typing.Union[typing.Callable[[T], bool], None]) -> "iter[T]":
+    def filter(
+        self, pred: typing.Union[typing.Callable[[T], bool], T, None] = None
+    ) -> "iter[T]":
         """Return an iterator containing only the elements for which pred
         returns True.
 
         If pred is None, return an iterator containing only elements that are
         truthy.
+
+        If pred is a T (and T is not a callable or None), return an iterator
+        containing only elements that compare equal to pred.
         """
+        if not callable(pred) and pred is not None:
+            pred = (lambda j: lambda i: i == j)(pred)
         return iter(filter(pred, self))
 
+    def find(
+        self, pred: typing.Union[typing.Callable[[T], bool], T, None] = None
+    ) -> typing.Optional[T]:
+        """Return the first element of self for which pred returns True.
+
+        If pred is None, return the first element which is truthy.
+
+        If pred is a T (and T is not a callable or None), return the first element
+        that compares equal to pred.
+
+        If no such element exists, return None.
+        """
+        if pred is None:
+            pred = lambda i: bool(i)
+        elif not callable(pred):
+            pred = (lambda j: lambda i: i == j)(pred)
+        for i in self:
+            if pred(i):
+                return i
+
     @typing.overload
     def reduce(self, func: typing.Callable[[T, T], T]) -> T:
         ...
@@ -346,8 +400,8 @@ class iter(typing.Generic[T]):
         initial is only usable on versions of Python equal to or greater than 3.8.
         """
         if initial is self._SENTINEL:
-            return iter(itertools.accumulate(func, self))
-        return iter(itertools.accumulate(func, self, initial))
+            return iter(itertools.accumulate(self, func))
+        return iter(itertools.accumulate(self, func, initial))  # type: ignore
 
     def foreach(self, func: typing.Callable[[T], typing.Any]) -> None:
         """Run func on every value in this iterator, immediately."""
@@ -480,6 +534,23 @@ class iter(typing.Generic[T]):
             return sum(self)
         return sum(self, initial)
 
+    @typing.overload
+    def prod(self) -> T:
+        ...
+
+    @typing.overload
+    def prod(self, initial: T) -> T:
+        ...
+
+    def prod(self, initial=_SENTINEL):
+        """Return the product of all elements in this iterator.
+
+        If initial is provided, it is used as the initial value.
+        """
+        if initial is self._SENTINEL:
+            return math.prod(self)
+        return math.prod(self, start=initial)
+
     def sorted(
         self,
         key: typing.Union[typing.Callable[[T], U], None] = None,
@@ -557,8 +628,18 @@ class iter(typing.Generic[T]):
             )
         )
 
+    def enumerate(self, start: int = 0) -> "iter[typing.Tuple[int, T]]":
+        """Return an iterator over the elements of this iterator, paired with
+        their index, starting at start.
+        """
+        return iter(enumerate(self, start))
+
+    def count(self) -> int:
+        """Consume this iterator and return the number of elements it contained."""
+        return self.map(lambda _: 1).sum()
+
     def __repr__(self) -> str:
-        return f"Smart({self.it!r})"
+        return f"iter({self.it!r})"
 
 
 @functools.wraps(builtins.range)
@@ -626,7 +707,7 @@ def tail_call(func: typing.Callable[P, U]) -> typing.Callable[P, U]:
     """
 
     @functools.wraps(func)
-    def wrapped(*args, **kwargs):
+    def wrapped(*args: P.args, **kwargs: P.kwargs):
         f = sys._getframe()
         if f.f_back and f.f_back.f_back and f.f_back.f_back.f_code == f.f_code:
             raise TailRecursionDetected(args, kwargs)
@@ -635,7 +716,275 @@ def tail_call(func: typing.Callable[P, U]) -> typing.Callable[P, U]:
                 try:
                     return func(*args, **kwargs)
                 except TailRecursionDetected as e:
-                    args = e.args
-                    kwargs = e.kwargs
+                    args = e.args  # type: ignore
+                    kwargs = e.kwargs  # type: ignore
+        raise Exception("unreachable")
 
     return wrapped
+
+
+LetterRow = typing.Tuple[
+    bool,
+    bool,
+    bool,
+    bool,
+    bool,
+]
+Letter = typing.Tuple[
+    LetterRow,
+    LetterRow,
+    LetterRow,
+    LetterRow,
+    LetterRow,
+    LetterRow,
+]
+
+
+def encode_letter(dots: Letter) -> int:
+    """Encode a matrix of dots to an integer for efficient
+    storage and lookup. Not expected to be used outside of
+    this module and contributions to the lookup table.
+
+    The matrix of dots should be 6 tall and 5 wide.
+    """
+    # Letters are 4 dots wide; the 5th column should always be empty.
+    # This function assumes that input is not malformed; any dots in the
+    # 5th column are treated as if they are in the 1st column of the next
+    # row.
+    # If something includes the 5th column it is malformed, but this
+    # function will not check.
+    out = 0
+    for y, row in enumerate(dots):
+        for x, dot in enumerate(row):
+            if dot:
+                out |= 1 << (x + 4 * y)
+    return out
+
+
+LETTERS: typing.Dict[int, str] = {
+    # todo: fill in this lookup table
+    0: " ",
+    10090902: "A",
+    7968663: "B",
+    6885782: "C",
+    15800095: "E",
+    1120031: "F",
+    15323542: "G",
+    10067865: "H",
+    14959694: "I",
+    6916236: "J",
+    9786201: "K",
+    15798545: "L",
+    6920598: "O",
+    1145239: "P",
+    9795991: "R",
+    7889182: "S",
+    6920601: "U",
+    4475409: "Y",
+    15803535: "Z",
+}
+
+
+def decode_letter(dots: Letter) -> str:
+    """Decode a matrix of dots to a single letter.
+
+    The matrix of dots should be 6 tall and 5 wide.
+    """
+    encoded = encode_letter(dots)
+    if encoded not in LETTERS:
+        print("Unrecognised letter:", encoded)
+        for row in dots:
+            for dot in row:
+                print(" #"[dot], end="")
+            print()
+        print("Please consider contributing this to the lookup table:")
+        print("https://github.com/starwort/aoc_helper")
+        return "?"
+    return LETTERS[encoded]
+
+
+def decode_text(dots: typing.List[typing.List[bool]]) -> str:
+    """Decode a matrix of dots to text.
+
+    The matrix of dots should be 6 tall and 5n - 1 wide.
+    """
+    broken_rows = [list(chunk_default(row, 5, False)) for row in dots]
+    letters = list(zip(*broken_rows))
+    out = []
+    for letter in letters:
+        out.append(decode_letter(letter))
+    if "?" in out:
+        # prevent submitting malformed output
+        raise ValueError("Unrecognised letter")
+    return "".join(out)
+
+
+def _default_classifier(char: str) -> int:
+    if char in "0123456789":
+        return int(char)
+    elif char in ".#":
+        return ".#".index(char)
+    else:
+        raise ValueError(f"Could not classify {char}. Please use a custom classifier.")
+
+
+class Grid(typing.Generic[T]):
+    data: list[list[T]]
+
+    def __init__(self, data: list[list[T]]) -> None:
+        self.data = data
+
+    @typing.overload
+    @classmethod
+    def from_string(cls, data: str) -> "Grid[int]":
+        ...
+
+    @typing.overload
+    @classmethod
+    def from_string(cls, data: str, classify: typing.Callable[[str], T]) -> "Grid[T]":
+        ...
+
+    @classmethod
+    def from_string(cls, data: str, classify=_default_classifier):  # type: ignore
+        """Create a grid from a string (e.g. a puzzle input).
+
+        Can take a classifier to use a custom classification. The default will
+        map numbers from 0 to 9 to themselves, and . and # to 0 and 1 respectively.
+        """
+        return Grid(list(data.splitlines()).mapped(lambda i: list(i).mapped(classify)))
+
+    def dijkstras(
+        self,
+        start: typing.Optional[typing.Tuple[int, int]] = None,
+        end: typing.Optional[typing.Tuple[int, int]] = None,
+    ) -> int:
+        """Use Dijkstra's algorithm to find the best path from
+        start to end, and return the total cost.
+
+        start defaults to the top left, and end defaults to the bottom right.
+        """
+        to_visit = []
+        heappush(to_visit, (0, start or (0, 0)))
+        visited = set()
+        if end is None:
+            target = len(self.data[0]) - 1, len(self.data) - 1
+        else:
+            target = end
+
+        while True:
+            cost, (x, y) = heappop(to_visit)
+            if (x, y) in visited:
+                continue
+            if (x, y) == target:
+                return cost
+            visited.add((x, y))
+            if x > 0:
+                heappush(to_visit, (cost + self.data[y][x - 1], (x - 1, y)))
+            if x < len(self.data[0]) - 1:
+                heappush(to_visit, (cost + self.data[y][x + 1], (x + 1, y)))
+            if y > 0:
+                heappush(to_visit, (cost + self.data[y - 1][x], (x, y - 1)))
+            if y < len(self.data) - 1:
+                heappush(to_visit, (cost + self.data[y + 1][x], (x, y + 1)))
+
+    def neighbours(self, x: int, y: int) -> list[T]:
+        """Return the neighbours of a point in the grid (but not the point itself).
+
+        Examples below:
+        - A is the point (x, y)
+        - * are points returned
+        - . are other points in the grid
+
+        ...........
+        ..***......
+        ..*A*......
+        ..***......
+
+        A*.........
+        **.........
+        ...........
+        ...........
+        """
+        return (
+            irange(max(y - 1, 0), min(y + 1, len(self.data) - 1))
+            .map(
+                lambda y_: irange(max(x - 1, 0), min(x + 1, len(self.data[0]) - 1))
+                .filter(lambda x_: (x, y) != (x_, y_))
+                .map(lambda x: self.data[y_][x])
+            )
+            .flatten(False)
+        ).collect()
+
+    def orthogonal_neighbours(self, x: int, y: int) -> list[T]:
+        """Return the orthogonal neighbours of a point in the grid (but not the
+        point itself).
+
+        Examples below:
+        - A is the point (x, y)
+        - * are points returned
+        - . are other points in the grid
+
+        ...........
+        ...*.......
+        ..*A*......
+        ...*.......
+
+        A*.........
+        *..........
+        ...........
+        ...........
+        """
+        rv = list()
+        if x > 0:
+            rv.append(self.data[y][x - 1])
+        if x < len(self.data[0]) - 1:
+            rv.append(self.data[y][x + 1])
+        if y > 0:
+            rv.append(self.data[y - 1][x])
+        if y < len(self.data) - 1:
+            rv.append(self.data[y + 1][x])
+        return rv
+
+    def deepcopy(self) -> "Grid[T]":
+        return Grid(self.data.deepcopy())
+
+    def __getitem__(self, index: int) -> list[T]:
+        return self.data[index]
+
+
+def dijkstras(
+    grid: typing.List[typing.List[int]],
+    start: typing.Optional[typing.Tuple[int, int]] = None,
+    end: typing.Optional[typing.Tuple[int, int]] = None,
+) -> int:
+    """Use Dijkstra's algorithm to find the best path from
+    start to end, and return the total cost.
+
+    start defaults to the top left, and end defaults to the bottom right.
+    """
+    return Grid(list(grid).mapped(list)).dijkstras(start, end)
+
+
+class PrioQueue(typing.Generic[T], typing.Iterator[T], typing.Iterable[T]):
+    _data: builtins.list[T]
+
+    def __init__(self, data: builtins.list[T]) -> None:
+        self._data = data
+        heapify(self._data)
+
+    def __next__(self) -> T:
+        if not self._data:
+            raise StopIteration
+        return heappop(self._data)
+
+    def __iter__(self):
+        return self
+
+    def __bool__(self) -> bool:
+        return bool(self._data)
+
+    def next(self) -> T:
+        return next(self)
+
+    def push(self, val: T) -> None:
+        heappush(self._data, val)
